// AForge Neural Net Library
//
// Copyright  Andrew Kirillov, 2005-2006
// andrew.kirillov@gmail.com
//

namespace NeuralNetworkPro.Activation_Functions
{
	using System;

	/// <summary>
	/// Sigmoid activation function
	/// </summary>
	///
	/// <remarks>The class represents sigmoid activation function with
	/// the next expression:<br />
	/// <code>
	///                1
	/// f(x) = ------------------
	///        1 + exp(-alpha * x)
	///
	///           alpha * exp(alpha * x )
	/// f'(x) = ---------------------------- = alpha * f(x) * (1 - f(x))
	///           (1 + exp(alpha * x))^2
	/// </code>
	/// Output range of the function: <b>[0, 1]</b><br /><br />
	/// Functions graph:<br />
	/// <img src="sigmoid.bmp" width="242" height="172" />
	/// </remarks>
    [Serializable]
	public class SigmoidFunction : IActivationFunction
	{
		// sigmoid's alpha value
		private double alpha = 2;

		/// <summary>
		/// Sigmoid's alpha value
		/// </summary>
		/// 
		/// <remarks>The value determines steepness of the function. Default value: <b>2</b>.
		/// </remarks>
		public double Alpha
		{
			get { return alpha; }
			set { alpha = value; }
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="SigmoidFunction"/> class
		/// </summary>
		public SigmoidFunction( ) { }

		/// <summary>
		/// Initializes a new instance of the <see cref="SigmoidFunction"/> class
		/// </summary>
		/// 
		/// <param name="alpha">Sigmoid's alpha value</param>
		public SigmoidFunction( double alpha )
		{
			this.alpha = alpha;
		}


		/// <summary>
		/// Calculates function value
		/// </summary>
		///
		/// <param name="x">Function input value</param>
		/// 
		/// <returns>Function output value, <i>f(x)</i></returns>
		///
		/// <remarks>The method calculates function value at point <b>x</b>.</remarks>
		///
		public double Function( double x )
		{
			return ( 1 / ( 1 + Math.Exp( -alpha * x ) ) );
		}

		/// <summary>
		/// Calculates function derivative
		/// </summary>
		/// 
		/// <param name="x">Function input value</param>
		/// 
		/// <returns>Function derivative, <i>f'(x)</i></returns>
		/// 
		/// <remarks>The method calculates function derivative at point <b>x</b>.</remarks>
		///
		public double Derivative( double x )
		{
            var y = Function(x);
            //var val = (alpha * Math.Exp(alpha * x)) / Math.Pow(Math.Exp(alpha * x) + 1, 2);
            //if (double.IsNaN(val))
            //    Console.WriteLine("Sigmoid pls");
            return alpha * y * (1 - y);
		}

        public override string ToString()
        {
            return Name + ":" + alpha;
        }

        public string Name
        {
            get { return "SigmoidFunction"; }
        }

        public static readonly string FunctionName = "SigmoidFunction"; 
	}
}
