﻿module UnfoldMonad
open System
open Microsoft.Xna.Framework
open Microsoft.Xna.Framework.Graphics
open Microsoft.Xna.Framework.Content
open System.Text.RegularExpressions
open Microsoft.FSharp.Quotations
open Component


type Unfold<'a,'r> = ('r * Time) -> Step<'a,'r>
and Step<'a,'r> = Done of 'a | Next of Unfold<'a,'r>



type UnfoldBuilder<'r>() =
  member this.Delay f = f()
  member this.Return(x:'a) :Unfold<'a,'r> =
    fun s -> Done x
  member this.Bind(p:Unfold<'a,'r>,k:'a->Unfold<'b,'r>):Unfold<'b,'r> =
    fun s ->
      match p s with
      |Done x -> Next(k x)
      |Next p' -> Next(this.Bind(p',k))
  member this.Combine(p:Unfold<'a,'r>,r:Unfold<'b,'r>):Unfold<'b,'r> =
    fun s ->
      match p s with
      |Done x -> Next r
      |Next p' -> Next(this.Combine(p',r))
  member this.For(items:seq<'a>,body:'a->Unfold<Unit,'r>):Unfold<Unit,'r> =
    fun s ->
      if items |> Seq.length = 0 then
        Done ()
      else
        let hd,tl = items |> Seq.head,items |> Seq.skip 1
        Next(this.Combine(body hd,this.For(tl,body)))
  member this.While(cond : Unit->bool,body:Unfold<Unit,'r>):Unfold<Unit,'r> =
    fun s ->
      if cond () = false then
        Done ()
      else
        match body s with
        |Done () -> Next(this.While(cond,body))
        |Next body' -> Next(this.Combine(body',this.While(cond,body)))
  member this.Zero() = this.Return()
  member this.ReturnFrom p = p


type Script<'a> = Unfold<'a,RenderState>
let unfold = UnfoldBuilder<RenderState>()



let set_render_texture texture p c s : Script<Unit> =
  fun (r,t) -> Done(r.Texture <- Some(texture,p,s,c))
let remove_render_texture : Script<Unit> =
  fun (r,t) -> Done(r.Texture <- None)

let set_render_text text p c s : Script<Unit> =
  fun (r,t) -> Done(r.Text <- Some(text,p,s,c))
let remove_render_text : Script<Unit> =
  fun (r,t) -> Done(r.Text <- None)

let time (n,t) = Done t

let wait (max_dt:float32) =
  let rec wait t0 =
    unfold{
      let! t = time
      let dt = t.T - t0.T
      if dt < max_dt then
        return! wait t0
    }
  unfold {
    let! t0 = time
    return! wait t0
  }
let wait_and_print max_dt msg (color:Color) text_center text_size =
  let rec wait t0 =
    unfold{
      let! t = time      
      let dt = t.T - t0.T
      do! set_render_text (msg+" "+(int (max_dt - dt)).ToString()) text_center color text_size
      if dt < max_dt then
        return! wait t0
    }
  unfold {
    let! t0 = time
    return! wait t0
  }
let wait_and_print_without_time max_dt msg (color:Color) text_center text_size =
  let rec wait t0 =
    unfold{
      let! t = time      
      let dt = t.T - t0.T
      do! set_render_text msg text_center color text_size
      if dt < max_dt then
        return! wait t0
    }
  unfold {
    let! t0 = time
    return! wait t0
  }

let rec forever_ (s : Unfold<Unit,_>) : Unfold<Unit,_> =
  unfold{
    do! s
    return! forever_ s
  }

let rec set_render_texts (msgLst:string list) (interval:float32) text_center text_size =
  unfold{
    do! match msgLst with
        |[] -> unfold {return ()}
        |msg::lst -> 
          unfold{
            do! set_render_text msg text_center Color.White text_size
            do! wait interval
            do! set_render_texts lst interval text_center text_size
          }
  }
let rec while_ (c : Script<bool>) (b:Script<Unit>) : Script<Unit> =
  unfold{
    let! c_v = c
    if c_v then
      do! b
      return! while_ c b
    else
      return ()
  }

let ignore_ (p:Script<'a>) : Script<Unit> =
  unfold{
    let! _ = p
    return ()
  }

let rec parallel_ (p:Script<'a>) (r:Script<'b>) : Script<'a * 'b> =
  fun s ->
    match p s, r s with
    | Done x, Done y -> Done(x,y)
    | Next p', Next r' -> Next(parallel_ p' r')
    | Next p', Done y -> Next(parallel_ p' (unfold.Return(y)))
    | Done x, Next r' -> Next(parallel_ (unfold.Return(x)) r')

let rec parallel_many_ = 
  function
  | [] -> unfold{return ()}
  | (t:Script<'a>) :: (ts:Script<'a> list) -> parallel_ t (parallel_many_ ts) |> ignore_

type Either<' a,' b> = Left of ' a | Right of ' b
let rec parallel_first_ (p:Script<'a>) (r:Script<'b>) : Script<Either<'a, 'b>> =
  fun s ->
    match p s, r s with
    | Done x, _ -> Done(Left x)
    | _, Done y -> Done(Right y)
    | Next p', Next r' -> Next(parallel_first_ p' r')

let rec atomic_ (p:Script<'a>) : Script<'a> =
  fun s ->
    match p s with
    | Done x -> Done x 
    | Next p' -> atomic_ p' s

let not_ (p:Script<bool>) : Script<bool> =
  unfold{
    let! res = p
    return res |> not
  }

let and_ (a:Script<bool>) (b:Script<bool>) : Script<bool> =
  unfold{
    let! a = a
    let! b = b
    return a && b
  }

let or_ (a:Script<bool>) (b:Script<bool>) : Script<bool> =
  unfold{
    let! a = a
    let! b = b
    return a || b
  }

let rec run_ (p:Script<'a>) =
  fun s ->
    match p s with
    | Done x -> x 
    | Next p' -> run_ p' s

let wait_while_ c = while_ c (unfold{ return () })

let show_for t p c s dt =
  unfold{
    do! set_render_text t p c s
    do! wait dt
    do! remove_render_text
  }

let iter_ f l =
  unfold{
    for x in l do f x
  }

let update_current_time =
    fun (current_time:Time ref) (dt:float32)->
        current_time := {T = current_time.Value.T + dt; DT = dt}
let update_script = 
  fun (script_ref:Unfold<Unit,RenderState> ref) (items:Items) (time_ref:(Time ref)) ->
    script_ref := match script_ref.Value (items.renderer,time_ref.Value) with
                  |Done res ->             
                    (fun _ -> Done res)
                  | Next s' -> s'
let draw_script =
  fun (game:Game) (items:Items) ->
    items.sprite_batch.Begin(SpriteSortMode.Immediate,
            BlendState.AlphaBlend, SamplerState.LinearClamp, 
            DepthStencilState.None, null, null, 
            Matrix.CreateScale(items.screen_size.X / 100.0f,
                                items.screen_size.Y / 100.0f, 1.0f))

    let draw_string_centered (f:SpriteFont) (t:String) (p:Vector2) (c:Color) (s:Vector2) =
      let t_size = f.MeasureString(t)
      let t_center = t_size * 0.5f
      let scale = 100.0f / items.screen_size.X // min (s.X / t_size.X) 1.0f
      items.sprite_batch.DrawString(f, t, p, c, 0.0f, t_center, scale * 0.9f, SpriteEffects.None, 0.0f)
    
    match items.renderer.Text with
    | Some(t,p,s,c) -> 
      let r = Rectangle((p.X - s.X * 0.5f) |> int, (p.Y - s.Y * 0.5f) |> int, 
                        (s.X) |> int, (s.Y) |> int)
      items.sprite_batch.Draw(items.text_bar, r, Color.Black)
      draw_string_centered (items.sprite_font) t p c s
    | _ -> ()

    match items.renderer.Texture with
    | Some(t,p,s,c) -> 
      let r = Rectangle((p.X - s.X * 0.5f) |> int, (p.Y - s.Y * 0.5f) |> int, 
                        (s.X) |> int, (s.Y) |> int)
      items.sprite_batch.Draw(game.Content.Load(t), r, Color.White)
    | _ -> ()

    items.sprite_batch.End()

